#include <libgen.h>
#include <assert.h>
#include <stdlib.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <sys/select.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <fcntl.h>

#include "buf.h"


int debugmode = 0;

growbuf_t stdoutbuf;
growbuf_t stderrbuf;

typedef struct pipe pipe_t;

typedef struct target target_t;
struct target {
	pipe_t *p;        /* primary output of this target */
	target_t *copy;   /* copy this target onto another target */
};

target_t *addtarget(target_t **oldtarget, pipe_t *p)
{
	target_t *n = malloc(sizeof *n);
	if(!n) abort();
	n->p = p;
	n->copy = NULL;

	if(*oldtarget) {
		target_t *tgp = *oldtarget;
		while(tgp->copy) tgp = tgp->copy;
		tgp->copy = n;
	} else {
		*oldtarget = n;
	}

	return n;
}

typedef struct {
	pid_t pid;
	char **argv;

	int inp[2];    /* pipe for writing on process stdin */
	int outp[2];   /* pipe for reading from process stdout */
	int errp[2];   /* pipe for reading from process stderr */

	int incnt;     /* outputs pointed to this stdin */
	int outcnt;    /* pipes from this stdout */
	int errcnt;    /* pipes from this stderr */

	growbuf_t wbuf; /* write buffer */

	lnbuf_t lnout; /* line buffer for stdout */
	lnbuf_t lnerr; /* line buffer for stderr */

	target_t *tgtout; /* target of stdout */
	target_t *tgterr; /* target of stderr */

	int alive;     /* process alive */
	int inalive;   /* stdin alive */
	int outalive;  /* stdout alive */
	int erralive;  /* stderr alive */
} proc_t;

proc_t *procs = NULL;
int nproc = 0;
int liveproc = 0;

proc_t * addproc()
{
	nproc ++;
	procs = realloc(procs, nproc * sizeof *procs);
	if(!procs) abort();

	proc_t *n = procs + (nproc - 1);
	n->pid = 0;
	n->argv = NULL;

	n->inp[0] = -1;  n->inp[1] = -1;
	n->outp[0] = -1; n->outp[1] = -1;
	n->errp[0] = -1; n->errp[1] = -1;

	n->incnt = 0;
	n->outcnt = 0;
	n->errcnt = 0;

	n->tgtout = NULL;
	n->tgterr = NULL;

	n->alive = 1;
	n->inalive = 1;
	n->outalive = 1;
	n->erralive = 1;

	growbuf_init(&n->wbuf);

	return n;
}

struct pipe {
	proc_t *inproc;  /* input process, NULL if our stdin */
	int inerr;       /* whether to use stderr of input process */

	proc_t *outproc; /* output process, NULL if our stdout */
	int outerr;      /* whether to use our stderr */

	char *filter;    /* prefix filter */
	int cut;         /* cut N chars off the beginning of every input line */
	int lfilter;     /* length of filter in chars */
	int term;        /* terminate pipe chain if filter matches */

	char *prefix;    /* prepend string to every output line */

	int direct;      /* is this a direct pipe? */
	int alive;
};

pipe_t *pipes;
int npipe=0;

pipe_t * addpipe()
{
	npipe ++;
	pipes = realloc(pipes, npipe * sizeof *pipes);
	if(!pipes) abort();

	pipe_t *n = pipes + (npipe - 1);
	n->inproc = NULL;
	n->inerr = 0;
	n->outproc = NULL;
	n->outerr = 0;
	n->filter = NULL;
	n->cut = 0;
	n->prefix = NULL;
	n->direct = 0;
	n->term = 0;
	n->alive = 1;
	return n;
}

/* close all stored FDs */
void closeall()
{
#define CLOSE(x) if((x)>2 ) close(x);
	int i;
	proc_t *p;
	for(i=0, p=procs; i<nproc; i++, p++) {
		CLOSE(p->inp[0]); CLOSE(p->inp[1]);
		CLOSE(p->outp[0]); CLOSE(p->outp[1]);
		CLOSE(p->errp[0]); CLOSE(p->errp[1]);
	}
}

/* parse non-negative integer from string
 * if invalid, returns -1 */
int getint(const char *str)
{
	if(!str[0]) return -1;

	char *end = NULL;
	int v = (int) strtol(str, &end, 0);
	if(*end) return -1;

	return v;
}

/* shutdown */
void killit()
{
	signal(SIGTERM, SIG_IGN);

	int i;
	proc_t *p;
	for(i=0, p=procs; i<nproc; i++, p++) {
		if(!p->alive) continue;
		killpg(p->pid, SIGTERM);
	}

	exit(1);
}

/* handle sigterm/sigint */
void sigtermhandler(int x)
{
	killit();
}

/* write contents of growbuf to an fd
 * returns nonzero on error */
int wrbuf(growbuf_t *b, int fd)
{
	if(!growbuf_hasdata(b)) return 0;
	int w = write(fd, b->cur, growbuf_hasdata(b));
	if(w < 0) return 1;
	growbuf_consume(b, w);
	return 0;
}

/* do a read from an fd
 * returns 1 on eof/error */
int doread(int fd, lnbuf_t *buf)
{
	static char b[2048];
	int r = read(fd, b, 2048);
	if(r < 1) return 1;
	lnbuf_write(buf, b, r);
	return 0;
}

/* we received a full line from some source */
void gotline(char *line, char *endline, target_t *tgt)
{
	pipe_t *p = tgt->p;
	char *wrline = line;

	/* apply filter and cut */
	if(p->filter) {
		*endline = 0;
		if(strncmp(line, p->filter, p->lfilter)) {
			*endline = '\n';
			goto cont;
		}
		if(p->cut) wrline += p->cut;
		*endline = '\n';
	}

	/* determine output buffer */
	growbuf_t *outbuf;
	if(p->outproc) {
		outbuf = &p->outproc->wbuf;
	} else {
		if(p->outerr) outbuf = &stderrbuf; else outbuf = &stdoutbuf;
	}

	/* plumb debug mode */
	if(debugmode) {
		pipe_t *p = tgt->p;
		char *sauce = p->inproc ? p->inproc->argv[0] : "stdin";
		char *drain = p->outproc ? p->outproc->argv[0] : (p->outerr ? "stderr" : "stdout");
		int xsauce = p->inproc ? (1 + p->inproc - procs) : 0;
		int xdrain = p->outproc ? (1 + p->outproc - procs) : 0;

		fprintf(stderr, "## line output %d:%s -> %d:%s, %d bytes: ", xsauce, sauce, xdrain, drain, endline - wrline + 1);
		fwrite(wrline, endline - wrline + 1, 1, stderr);
		fflush(stderr);
	}

	/* write prefix if any */
	if(p->prefix) growbuf_write(outbuf, p->prefix, strlen(p->prefix));

	/* write line */
	growbuf_write(outbuf, wrline, endline - wrline + 1);

	if(p->term) return;
cont:
	if(tgt->copy) gotline(line, endline, tgt->copy);
}

int directstdin=0; /* whether STDIN is a direct pipe */
int stdinactive=1; /* whether STDIN is active */

lnbuf_t stdinbuf;

int cntstdin=0;    /* reference count for STDIN */
int cntstdout=0;
int cntstderr=0;

target_t *stdintgt = NULL; /* STDIN targets */

/* set up input buffers */
void initinbufs()
{
	if(cntstdin > 0 && !directstdin && stdinactive)
		lnbuf_init(&stdinbuf, (lnproc_t) gotline, stdintgt);

	int i;
	proc_t *p;
	for(i=0, p=procs; i<nproc; i++, p++) {
		if(p->outalive && p->outcnt > 0)
			lnbuf_init(&p->lnout, (lnproc_t) gotline, p->tgtout);

		if(p->erralive && p->errcnt > 0)
			lnbuf_init(&p->lnerr, (lnproc_t) gotline, p->tgterr);
	}
}

/* mark pipes going from stdin as dead */
void stdindead()
{
	int i;
	pipe_t *q;

	if(!stdinactive) return;
	stdinactive=0;

	for(i=0, q=pipes; i<npipe; i++, q++) {
		if(!q->alive) continue;
		if(!q->inproc) {
			q->alive = 0;
			if(q->direct) continue;

			if(q->outproc) {
				if(q->outproc->alive) {
					assert(q->outproc->incnt > 0);
					q->outproc->incnt --;
					if(!q->outproc->incnt && !growbuf_hasdata(&q->outproc->wbuf)) {
						q->outproc->inalive = 0;
						close(q->outproc->inp[1]);
					}
				}
			} else {
				if(q->outerr) {
					assert(cntstderr > 0);
					cntstderr --;
					if(!cntstderr) close(2);
				} else {
					assert(cntstdout > 0);
					cntstdout --;
					if(!cntstdout) close(1);
				}
			}
		}
	}
}

/* mark pipes going from process as dead */
void processdeadpipes(proc_t *p, int out, int err)
{
	int i;
	pipe_t *q;

	if(out) {
		if(!p->outalive) out=0; else p->outalive=0;
	}
	if(err) {
		if(!p->erralive) err=0; else p->erralive=0;
	}

	if(!out && !err) return;

	/* mark relevant pipes as dead
	 * if an input runs out of pipes going to it, close input */

	for(i=0, q=pipes; i<npipe; i++, q++) {
		if(!q->alive) continue;

		if(q->inproc == p) {
			if(q->inerr && !err) continue;
			if(!q->inerr && !out) continue;

			q->alive = 0;
			if(q->direct) continue;

			if(q->outproc) {
				if(q->outproc->alive) {
					assert(q->outproc->incnt > 0);
					q->outproc->incnt --;
					if(!q->outproc->incnt && !growbuf_hasdata(&q->outproc->wbuf)) {
						q->outproc->inalive = 0;
						close(q->outproc->inp[1]);
					}
				}
			} else {
				if(q->outerr) {
					assert(cntstderr > 0);
					cntstderr --;
				} else {
					assert(cntstdout > 0);
					cntstdout --;
				}
			}
		}
	}
}

int keepalive=0;

void parseplumbcmd(int argc, char *argv[])
{
	/* parse plumb-mode cmd line */
	int i, pr;
	proc_t *p;
	pipe_t *q;

	if(argc < 2) {
		fprintf(stderr, "Usage: %s [-a] <processes...> <pipes...>\n", argv[0]);
		fprintf(stderr, "  -a   keep chain alive until all processes/inputs are gone\n\n");
		fprintf(stderr, "Process spec:\n");
		fprintf(stderr, "  + <program> <arg1> <arg2...> ;\n");
		fprintf(stderr, "  : <shell command>               same as + /bin/sh -c command ;\n");
		fprintf(stderr, "Pipe spec:\n");
		fprintf(stderr, "  <input> [-p|-P|-k|-K filterprefix] <output> [+p prependprefix]\n");
		fprintf(stderr, "     -p filter lines\n");
		fprintf(stderr, "     -P filter lines, cut prefix\n");
		fprintf(stderr, "     -k filter lines, end chain\n");
		fprintf(stderr, "     -K filter lines, cut prefix, end chain\n\n");
		fprintf(stderr, "Pipe input:\n");
		fprintf(stderr, "  0        %s stdin\n", argv[0]);
		fprintf(stderr, "  1..N     stdout of numbered process\n");
		fprintf(stderr, "  e1..N    stderr of numbered process\n\n");
		fprintf(stderr, "Pipe output:\n");
		fprintf(stderr, "  0        %s stdout\n", argv[0]);
		fprintf(stderr, "  e        %s stderr\n", argv[0]);
		fprintf(stderr, "  1..N     stdin of numbered process\n");
		exit(0);
	}

	/* parse processes */

	i=1;

	if(!strcmp(argv[i], "--debug")) {
		debugmode=1;
		i++;
	}

	if(!strcmp(argv[i], "-a")) {
		keepalive=1;
		i++;
	}

	for(; i<argc; i++) {
		/* + proc ... ; */
		if(!strcmp("+", argv[i])) {
			p = addproc();
			i ++;
			p->argv = argv+i;
			while(i < argc && strcmp(";", argv[i])) i++;
			if(i == argc) {
				fprintf(stderr, "Unterminated process (+) argument\n");
				exit(1);
			}
			argv[i] = NULL;
		} else if(!strcmp(":", argv[i])) {
			p = addproc();
			i ++;

			if(i == argc) {
				fprintf(stderr, "Missing argument after :\n");
				exit(1);
			}

			p->argv = malloc(4 * sizeof(*p->argv));
			if(!p->argv) abort();
			p->argv[0] = "/bin/sh";
			p->argv[1] = "-c";
			p->argv[2] = argv[i];
			p->argv[3] = NULL;
		} else break;
	}

	/* parse pipes */

	while(i<argc) {
		/* <pipein> <pipeout> */
		q = addpipe();

		if(argv[i][0] == 'e') {
			/* input: e<number>, process stderr */
			pr = getint(argv[i]+1);
			if(pr < 1 || pr > nproc) {
				fprintf(stderr, "Invalid pipe input spec: %s\n", argv[i]);
				exit(1);
			}
			q->inproc = procs+(pr-1);
			q->inerr = 1;
		} else {
			/* input: <number>, process stdout (or our stdin) */
			q->inerr = 0;
			pr = getint(argv[i]);
			if(pr < 0 || pr > nproc) {
				fprintf(stderr, "Invalid pipe input spec: %s\n", argv[i]);
				exit(1);
			} else if(pr == 0) {
				q->inproc = NULL;
			} else {
				q->inproc = procs+(pr-1);
			}
		}

		i ++;
		if(i == argc) {
			fprintf(stderr, "Pipe output spec missing\n");
			exit(1);
		}

		if(!strcmp("-p", argv[i]) || !strcmp("-P", argv[i]) || !strcmp("-k", argv[i]) || !strcmp("-K", argv[i])) {
			q->cut = !strcmp("-P", argv[i]) || !strcmp("-K", argv[i]);
			q->term = !strcmp("-k", argv[i]) || !strcmp("-K", argv[i]);

			i ++;
			if(i == argc) {
				fprintf(stderr, "Need prefix arg after -p|-P\n");
				exit(1);
			}
			q->filter = argv[i];
			q->lfilter = strlen(q->filter);
			if(q->cut) q->cut = q->lfilter;
			if(!q->filter[0]) q->filter = NULL;

			i ++;
			if(i == argc) {
				fprintf(stderr, "Pipe output spec missing\n");
				exit(1);
			}
		}

		if(!strcmp("e", argv[i])) {
			/* output is our stderr */
			q->outproc = NULL;
			q->outerr = 1;
		} else {
			q->outerr = 0;
			pr = getint(argv[i]);
			if(pr < 0 || pr > nproc) {
				fprintf(stderr, "Invalid pipe output spec: %s\n", argv[i]);
				exit(1);
			}
			if(pr == 0) {
				q->outproc = NULL;
			} else {
				q->outproc = procs+(pr-1);
			}
		}

		i ++;
		if(i < argc) {
			if(!strcmp("+p", argv[i])) {
				i ++;
				if(i == argc) {
					fprintf(stderr, "Need prefix arg after +p\n");
					exit(1);
				}
				q->prefix = argv[i];
				i ++;
			}
		}
	}
}

void parse69cmd(int argc, char *argv[])
{
	proc_t *p1;
	proc_t *p2;
	pipe_t *q;

	/* parse 69-mode cmd line */

	if(argc < 3 || argc > 4) {
		fprintf(stderr, "Usage: %s command1 command2 [-a]\n", argv[0]);
		exit(1);
	}

	char **nargv = malloc(8 * sizeof *nargv);
	if(!nargv) abort();
	nargv[0] = "/bin/sh";
	nargv[1] = "-c";
	nargv[2] = argv[1];
	nargv[3] = NULL;
	nargv[4] = "/bin/sh";
	nargv[5] = "-c";
	nargv[6] = argv[2];
	nargv[7] = NULL;

	p1 = addproc();
	p1->argv = nargv;
	p2 = addproc();
	p2->argv = nargv+4;

	q = addpipe();
	q->inproc = p1;
	q->inerr = 0;
	q->outproc = p2;
	q->outerr = 0;

	q = addpipe();
	q->inproc = p2;
	q->inerr = 0;
	q->outproc = p1;
	q->outerr = 0;

	if(argc == 4) {
		if(!strcmp("-a", argv[3])) {
			keepalive = 1;
		} else {
			fprintf(stderr, "Invalid command line argument %s\n", argv[3]);
			exit(1);
		}
	}
}

int main(int argc, char *argv[], char *envp[])
{
	int i;
	proc_t *p;
	pipe_t *q;

	growbuf_init(&stdoutbuf);
	growbuf_init(&stderrbuf);

	/* parse command line */

	char *execname = basename(strdup(argv[0]));
	if(!strcmp(execname, "69")) {
		parse69cmd(argc, argv);
	} else {
		parseplumbcmd(argc, argv);
	}

	/* add reference counts of inputs/outputs */

	for(i=0, q=pipes; i<npipe; i++, q++) {
		if(q->inproc) {
			if(q->inerr) {
				q->inproc->errcnt ++;
				addtarget(&q->inproc->tgterr, q);
			} else {
				q->inproc->outcnt ++;
				addtarget(&q->inproc->tgtout, q);
			}
		} else {
			cntstdin ++;
			addtarget(&stdintgt, q);
		}

		if(q->outproc) {
			q->outproc->incnt ++;
		} else {
			if(q->outerr) cntstderr++; else cntstdout++;
		}
	}

	/* calculate which pipes can be direct */
	for(i=0, q=pipes; i<npipe; i++, q++) {
		if(q->filter || q->prefix) continue;

		if(q->inproc) {
			if(q->inerr && q->inproc->errcnt != 1)
				continue;
			if(!q->inerr && q->inproc->outcnt != 1)
				continue;

			if(q->outproc) {
				if(q->outproc->incnt == 1)
					q->direct = 1;
			} else {
				if(q->outerr && cntstderr == 1)
					q->direct = 1;
				else if(!q->outerr && cntstdout == 1)
					q->direct = 1;
			}
		} else {
			if(q->outproc && q->outproc->incnt == 1 && cntstdin == 1) {
				q->direct = 1;
				directstdin=1;
			}
		}
	}

	/* generate direct pipes */
	for(i=0, q=pipes; i<npipe; i++, q++) {
		if(!q->direct) continue;

		/* when not both endpoints are processes, we just pass on our own fd */
		if(!(q->inproc && q->outproc)) {
			if(!q->inproc) {
				assert(q->outproc);
				q->outproc->inp[0] = 0;
				q->outproc->incnt = 0;
			} else {
				assert(q->inproc);
				if(q->inerr) {
					q->inproc->errp[1]=q->outerr ? 2 : 1;
					q->inproc->errcnt = 0;
				} else {
					q->inproc->outp[1]=q->outerr ? 2 : 1;
					q->inproc->outcnt = 0;
				}
			}
			continue;
		}

		int p[2];
		if(pipe(p) == -1) {
			perror("Failed to create pipe");
			return 1;
		}

		if(q->inerr) q->inproc->errp[1] = p[1]; else q->inproc->outp[1] = p[1];
		q->outproc->inp[0] = p[0];

		/* set refcounts to 0, no need to create a process pipe there */
		if(q->inerr) q->inproc->errcnt=0; else q->inproc->outcnt=0;
		q->outproc->incnt = 0;
	}

	/* generate process pipes */
	for(i=0, p=procs; i<nproc; i++, p++) {
		if(p->incnt > 0) {
			assert(p->inp[0] == -1 && p->inp[1] == -1);
			if(pipe(p->inp) == -1) {
				perror("Failed to create pipe");
				return 1;
			}

			int f = fcntl(p->inp[1], F_GETFL, 0);
			if(f == -1) {
				perror("fcntl GETFL fails");
				fflush(stderr);
			} else if(fcntl(p->inp[1], F_SETFL, f | O_NONBLOCK) == -1) {
				perror("fcntl SETFL fails");
				fflush(stderr);
			}
		}
		if(p->outcnt > 0) {
			assert(p->outp[0] == -1 && p->outp[1] == -1);
			if(pipe(p->outp) == -1) {
				perror("Failed to create pipe");
				return 1;
			}
		}
		if(p->errcnt > 0) {
			assert(p->errp[0] == -1 && p->errp[1] == -1);
			if(pipe(p->errp) == -1) {
				perror("Failed to create pipe");
				return 1;
			}
		}
	}

	/* set up polling */

	initinbufs();

	/* set up signal handling */
	signal(SIGINT, sigtermhandler);
	signal(SIGTERM, sigtermhandler);
	signal(SIGPIPE, SIG_IGN);

	/* fork off processes */
	for(i=0, p=procs; i<nproc; i++, p++) {
		p->pid = fork();
		if(p->pid == -1) {
			perror("Failed to fork");
			killit();
		}

		if(p->pid == 0) {
			/* we're the new process */

			/* don't close stderr by default, maybe there's some hope of
			 * usefully reporting errors */
			if(p->inp[0] != -1) dup2(p->inp[0], 0); else close(0);
			if(p->outp[1] != -1) dup2(p->outp[1], 1); else if(p->outp[1] != 2) close(1);
			if(p->errp[1] != -1) dup2(p->errp[1], 2);

			closeall();

			if(setpgrp() == -1) {
				perror("setpgrp failed");
				fflush(stderr);
			}

			execvp(p->argv[0], p->argv);
			fprintf(stderr, "Failed to start process %s\n%s\n", p->argv[0], strerror(errno));
			exit(1);
		} else {
			/* we're the parent */

			/* need to close the unnecessary pipe sides */
			if(p->inp[0] != -1) close(p->inp[0]);
			if(p->outp[1] != -1) close(p->outp[1]);
			if(p->errp[1] != -1) close(p->errp[1]);
		}

		liveproc ++;
	}

	fd_set rfds;
	fd_set wfds;
	int nofd = 0;

	/* main loop */
	for(;;) {
		int pstatus;
		pid_t rp = waitpid(-1, &pstatus, nofd ? 0 : WNOHANG);
		if(rp == -1 && errno != ECHILD) {
			perror("waitpid fails");
			killit();
		}
		if(rp > 0) {
			/* a process died, handle it */

			if(debugmode) fprintf(stderr, "## Child process with pid %d dies\n", rp);

			for(i=0, p=procs; i<nproc; i++, p++)
				if(p->pid == rp) break;
			if(debugmode && i != nproc) fprintf(stderr, "## Dead process: number %d: %s\n", i, p->argv[0]);

			if(!keepalive) killit();
			if(!WIFEXITED(pstatus)) killit();
			if(WEXITSTATUS(pstatus)) killit();

			if(i != nproc) {
				if(p->alive) {
					p->alive = 0;
					liveproc --;
				}
			}
		}

		/* do the polling */

		int hfd = -1;
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);

		if(cntstdin > 0 && !directstdin && stdinactive) {
			FD_SET(0, &rfds);
			hfd = 0;
		}

		if(growbuf_hasdata(&stdoutbuf)) {
			FD_SET(1, &wfds);
			hfd = 1;
		}

		if(growbuf_hasdata(&stderrbuf)) {
			FD_SET(2, &wfds);
			hfd = 2;
		}

		for(i=0, p=procs; i<nproc; i++, p++) {
			if(p->inalive && growbuf_hasdata(&p->wbuf)) {
				FD_SET(p->inp[1], &wfds);
				if(p->inp[1] > hfd) hfd = p->inp[1];
			}

			if(p->outalive && p->outcnt > 0) {
				FD_SET(p->outp[0], &rfds);
				if(p->outp[0] > hfd) hfd = p->outp[0];
			}

			if(p->erralive && p->errcnt > 0) {
				FD_SET(p->errp[0], &rfds);
				if(p->errp[0] > hfd) hfd = p->errp[0];
			}
		}

		if(hfd == -1) {
			if(!liveproc) killit();
			nofd = 1;
			continue;
		} else {
			nofd = 0;
		}

		struct timeval tv;
		tv.tv_sec = 1;
		tv.tv_usec = 0;
		int r = select(hfd+1, &rfds, &wfds, NULL, &tv);

		if(r < 0) {
			if(errno == EINTR) continue;
			perror("select fails");
			killit();
		}

		if(r == 0) continue;

		/* check stdin */
		if(FD_ISSET(0, &rfds)) {
			assert(cntstdin > 0 && !directstdin && stdinactive);
			if(doread(0, &stdinbuf)) {
				if(!keepalive) killit();
				close(0); stdindead();
			}
		}

		/* check stdout and stderr */
		if(FD_ISSET(1, &wfds)) {
			if(wrbuf(&stdoutbuf, 1)) {
				perror("error writing stdout");
				/* TODO */
				killit();
			} else {
				if(!cntstdout) {
					p->outalive = 0;
					close(1);
				}
			}
		}

		if(FD_ISSET(2, &wfds)) {
			if(wrbuf(&stderrbuf, 2)) {
				perror("error writing stderr");
				/* TODO */
				killit();
			} else {
				if(!cntstderr) {
					p->erralive = 0;
				}
			}
		}

		/* check process outputs */
		for(i=0, p=procs; r>0 && i<nproc; i++, p++) {
			if(p->inalive && growbuf_hasdata(&p->wbuf) && FD_ISSET(p->inp[1], &wfds)) {
				if(wrbuf(&p->wbuf, p->inp[1])) {
					perror("error writing process stdin");
					/* TODO */
					killit();
				} else {
					if(!p->incnt) {
						close(p->inp[1]);
						p->inalive = 0;
					}
				}
			}

			if(p->outalive && p->outcnt > 0 && FD_ISSET(p->outp[0], &rfds)) {
				if(doread(p->outp[0], &p->lnout)) {
					if(!keepalive) killit();
					processdeadpipes(p, 1, 0);
				}
			}

			if(p->erralive && p->errcnt > 0 && FD_ISSET(p->errp[0], &rfds)) {
				if(doread(p->errp[0], &p->lnerr)) {
					if(!keepalive) killit();
					processdeadpipes(p, 0, 1);
				}
			}
		}
	}

	return 0;
}
